import textwrap

from math import floor
from typing import Callable, Dict, Optional, Type

import colorama
from colorama import Fore, Back

import xappt

from xappt.models.parameter.model import Parameter


# noinspection PyMethodMayBeStatic
@xappt.register_plugin
class StdIO(xappt.BaseInterface):
    def __init__(self):
        super().__init__()

        self.prompt_dispatch: Dict[Type, Callable] = {
            int: self.prompt_int,
            bool: self.prompt_bool,
            str: self.prompt_str,
            list: self.prompt_str,
        }

        self._progress_started = None
        self._term_size = (80, 25)

        colorama.init(autoreset=True)

    def message(self, message: str):
        self.progress_end()
        print(message)

    def warning(self, message: str):
        self.progress_end()
        print(f"{Fore.YELLOW}WARNING: {message}")

    def error(self, message: str, *, details: Optional[str] = None):
        self.progress_end()
        print(f"{Fore.RED}ERROR: {message}")
        if details is not None and len(details):
            print(f"{Fore.RED}\n{details}\n")

    def ask(self, message: str) -> bool:
        choices = ("y", "n")
        while True:
            result = input(f"{message} ({'|'.join(choices)}) ")
            if result not in choices:
                print(f"Please enter {xappt.humanize_list(choices)}")
                continue
            break
        return choices.index(result) == 0

    def progress_start(self):
        if self._progress_started:
            self.progress_end()
        self._progress_started = True
        self._term_size = xappt.get_terminal_size()

    def progress_update(self, message: str, percent_complete: float):
        if not self._progress_started:
            raise RuntimeError("`progress_start` not called.")

        percent_complete = max(0.0, min(1.0, percent_complete))

        max_width = self._term_size[0]
        message = textwrap.shorten(message, width=max_width).ljust(max_width)

        progress_width = int(floor(max_width * percent_complete))
        message_head = message[:progress_width]
        message_tail = message[progress_width:]

        print(f"\r{Fore.BLACK}{Back.WHITE}{message_head}{Fore.RESET}{Back.RESET}{message_tail}", end="")

    def progress_end(self):
        if not self._progress_started:
            return
        print("")
        self._progress_started = False

    def invoke(self, plugin: xappt.BaseTool, **kwargs):
        for param in plugin.parameters():
            self.prompt_for_param(param)
        return plugin.execute(interface=self, **kwargs)

    # noinspection PyMethodMayBeStatic
    def prompt_default(self, param: Parameter) -> str:
        if param.default is not None:
            return f"{param.name} ({param.default}): "
        return f"{param.name}: "

    def prompt_int(self, param: Parameter) -> str:
        if param.choices is not None:
            if param.default is not None:
                choice_values = []
                for i, value in enumerate(param.choices):
                    if i == param.default:
                        choice_values.append(f'{value}*')
                    else:
                        choice_values.append(value)
            else:
                choice_values = param.choices
            return f"{param.name} ({'|'.join(choice_values)}): "
        return self.prompt_default(param)

    def prompt_str(self, param: Parameter) -> str:
        if param.choices is not None:
            choice_values = []
            for value in param.choices:
                if value == param.default:
                    choice_values.append(f'{value}*')
                else:
                    choice_values.append(value)
            return f"{param.name} ({'|'.join(choice_values)}): "
        return self.prompt_default(param)

    # noinspection PyMethodMayBeStatic
    def prompt_bool(self, param: Parameter) -> str:
        choices = "(y|n)"
        if param.default is not None:
            if param.default:
                choices = "(y*|n)"
            else:
                choices = "(y|n*)"
        return f"{param.name} {choices}: "

    def prompt_for_param(self, param: Parameter):
        prompt_fn = self.prompt_dispatch.get(param.data_type, self.prompt_default)
        prompt = prompt_fn(param)

        print("")
        if len(param.description):
            print(param.description)

        while True:
            value = input(prompt)
            if len(value) == 0:
                value = None
            try:
                param.value = param.validate(value)
            except BaseException as e:
                print(f"{Fore.RED}Error: {e}")
                print("Try again")
                print("")
            else:
                break



import os

from .command_runner import CommandRunner

from xappt.config import log as logger


def is_path_repository(path: str) -> bool:
    """ Check if a path contains a git repository. """
    return CommandRunner(cwd=path).run(("git", "rev-parse")).result == 0


def commit_id(path: str, *, short: bool = False) -> str:
    """ Fetch the current commit id (optionally the short version) from a git
     repository located at `path`. """
    command = ["git", "rev-parse"]
    if short:
        command.append("--short")
    command.append("HEAD")
    output = CommandRunner(cwd=path).run(command)
    if output.result == 0:
        return output.stdout.strip()
    raise RuntimeError(output.stderr)


def is_dirty(path: str) -> bool:
    cmd = CommandRunner(cwd=path)

    # are we a git repository?
    if not is_path_repository(path):
        logger.warning(f"Not a git repository: {path}")
        return True

    cmd.run(("git", "fetch", "origin"))

    # check for local change
    result = cmd.run(("git", "diff-index", "--quiet", "HEAD", "--"))
    if result.result != 0:
        logger.warning(f"Local changes detected: {path}")
        return True

    # check for remote change
    result = cmd.run(("git", "diff", "--quiet", "master", "origin/master"))
    if result.result != 0:
        logger.warning(f"Remote changes detected: {path}")
        return True

    return False


def clone(url: str, *, destination: str, branch: str = "master") -> str:
    """ Clone a remote repository to `destination` and check out `branch`. This
    also mimics the behavior of creating a subfolder based on the name of the
    repository. """
    logger.info("Cloning repository %s", url)
    logger.debug("Repository branch %s", branch)
    logger.debug("Repository destination %s", destination)

    name = os.path.splitext(os.path.basename(url))[0]
    repo_path = os.path.join(destination, name)
    git_command = ("git", "clone", "-b", branch, url, repo_path)
    output = CommandRunner().run(git_command, silent=True)
    if output.result == 0:
        return repo_path
    raise RuntimeError(output.stderr)



#!/usr/bin/env python3

import argparse
import os
import sys

from collections import defaultdict
from itertools import chain
from typing import DefaultDict, List

import colorama
from colorama import Fore

import xappt

from xappt.models.parameter import convert


def build_parser() -> argparse.ArgumentParser:
    interface_list = [i[0] for i in xappt.plugin_manager.registered_interfaces()]
    default_interface_name = os.environ.get(xappt.INTERFACE_ENV, xappt.INTERFACE_DEFAULT)

    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--version', action='store_true',
                        help='Display the version number and build')
    parser.add_argument('-l', '--list', action='store_true',
                        help='List all of the discovered plugins')
    parser.add_argument('-i', '--interface', choices=interface_list, default=default_interface_name,
                        help='Specify the name of the default user interface. '
                             f'This can also be done by setting the environment variable {xappt.INTERFACE_ENV}')
    parser.add_argument('--ui', action='store_true', dest="interactive",
                        help='Run the specified tool interactively using the passed command line arguments, '
                             'otherwise it will be invoked with the default interface.')

    subparsers = parser.add_subparsers(help="Sub command help", dest='command')

    for plugin_name, plugin_class in xappt.plugin_manager.registered_tools():
        plugin_parser = subparsers.add_parser(plugin_class.name(), help=plugin_class.help())
        for parameter in plugin_class.class_parameters():
            setup_args = parameter.param_setup_args
            args, kwargs = convert.to_argument_dict(setup_args)
            plugin_parser.add_argument(*args, **kwargs)

    return parser


def list_all_plugins():
    colorama.init(autoreset=True)

    plugin_list: DefaultDict[str, List[str]] = defaultdict(list)

    for _, plugin_class in chain(xappt.plugin_manager.registered_interfaces(), xappt.plugin_manager.registered_tools()):
        collection = plugin_class.collection()
        plugin_list[collection].append(f"{Fore.LIGHTBLUE_EX}{plugin_class.name()} "
                                       f"{Fore.WHITE}({plugin_class.help() or 'No help text'})")

    for collection in sorted(plugin_list.keys()):
        print(f"{Fore.GREEN}{collection}")
        for plugin in sorted(plugin_list[collection]):
            print(f"    {plugin}")


def cli_main(*argv) -> int:
    parser = build_parser()
    options = parser.parse_args(args=argv)

    if options.version:
        print(f"xappt {xappt.version_str}")
        return 0

    if options.list:
        list_all_plugins()
        return 0

    os.environ[xappt.INTERFACE_ENV] = options.interface

    if options.command is not None:
        if options.interactive:
            interface = xappt.plugin_manager.get_interface()
        else:
            interface = xappt.plugin_manager.get_interface("stdio")

        tool_class = xappt.plugin_manager.get_tool_plugin(options.command)
        tool_init_kwargs = options.__dict__.copy()
        tool_init_kwargs.pop('interface')
        tool_instance = tool_class(interface=interface, **tool_init_kwargs)

        if options.interactive:
            interface.invoke(tool_instance)
        else:
            tool_instance.validate()
            return tool_instance.execute()
    else:
        parser.print_help()

    return 0


def entry_point() -> int:
    xappt.discover_plugins()
    return cli_main(*sys.argv[1:])


if __name__ == '__main__':
    sys.exit(entry_point())



import os
import subprocess

from typing import Optional, Tuple


def get_terminal_size() -> Tuple[int, int]:
    import platform
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        tuple_xy = 80, 25  # default value
    return tuple_xy


# noinspection PyBroadException,SpellCheckingInspection
def _get_terminal_size_windows() -> Optional[Tuple[int, int]]:
    import struct
    try:
        from ctypes import windll, create_string_buffer
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            (bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx, maxy) = \
                struct.unpack("hhhhHhhhhhh", csbi.raw)
            sizex = right - left + 1
            sizey = bottom - top + 1
            return sizex, sizey
    except BaseException:
        pass


# noinspection PyBroadException,SpellCheckingInspection
def _get_terminal_size_tput() -> Optional[Tuple[int, int]]:
    import shlex
    try:
        cols = int(subprocess.check_call(shlex.split('tput cols')))
        rows = int(subprocess.check_call(shlex.split('tput lines')))
        return cols, rows
    except BaseException:
        pass


# noinspection PyBroadException
def _get_terminal_size_linux() -> Optional[Tuple[int, int]]:
    import struct

    # noinspection PyShadowingNames,PyBroadException,SpellCheckingInspection
    def ioctl_gwinsz(fd):
        try:
            import fcntl
            import termios
            cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, b'1234'))
            return cr
        except BaseException:
            pass

    cr = ioctl_gwinsz(0) or ioctl_gwinsz(1) or ioctl_gwinsz(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_gwinsz(fd)
            os.close(fd)
        except BaseException:
            pass
    if not cr:
        try:
            cr = (os.environ['LINES'], os.environ['COLUMNS'])
        except BaseException:
            return None
    return int(cr[1]), int(cr[0])



import logging
import os

from xappt.constants import APP_NAME

log = logging.getLogger(APP_NAME)
log.setLevel(logging.WARN)
sh = logging.StreamHandler()
formatter = logging.Formatter("[%(levelname)8s] %(name)s : %(message)s")
sh.setFormatter(formatter)
sh.setLevel(logging.WARN)
log.addHandler(sh)

if os.environ.get("XAPPT_DEBUG", "0") != "0":
    sh.setLevel(logging.DEBUG)
    log.setLevel(logging.DEBUG)



import enum
import fnmatch
import os
import random
import re
import string

from typing import Generator, Sequence, Union

RAND_CHARS = string.ascii_letters + string.digits


class UniqueMode(enum.Enum):
    RANDOM = 0
    INTEGER = 1


def find_files(path: str, patterns: Union[str, Sequence], *, recursive: bool = False) -> Generator[str, None, None]:
    """ Search `path` for files, optionally *recursively*, and yield files that match `pattern`.
    Note that `pattern` expects an `fnmatch` compatible pattern (e.g. `*.py`), or a list/tuple of patterns.
    """
    if isinstance(patterns, str):
        patterns = [patterns]
    pattern_regex = re.compile('|'.join([fnmatch.translate(p) for p in patterns]))

    if recursive:
        for root, dirs, files in os.walk(path):
            for f in files:
                if pattern_regex.match(f):
                    yield os.path.join(root, f)
    else:
        for item in os.scandir(path):
            if pattern_regex.match(item.name):
                yield item.path


def get_unique_name(path: str, *, mode: UniqueMode = UniqueMode.RANDOM, **kwargs) -> str:
    """ Generate a unique file name by adding either random characters or sequential integers.

    >>> file_path = "/path/to/file.txt"
    >>> my_delimiter = "-"
    >>> unique_path = get_unique_name(file_path, mode=UniqueMode.RANDOM, length=8, force=True, delimiter=my_delimiter)
    >>> len(unique_path) == len(file_path) + 8 + len(my_delimiter)
    True
    >>> unique_path = get_unique_name(file_path, mode=UniqueMode.INTEGER, length=3, force=True, delimiter=my_delimiter)
    >>> len(unique_path) == len(file_path) + 3 + len(my_delimiter)
    True
    >>> unique_path.endswith(f"file{my_delimiter}001.txt")
    True

    """
    delimiter = kwargs.get('delimiter', "-")
    max_iterations = kwargs.get("max_iterations", 9999)
    force_unique = kwargs.get("force", False)

    if mode == UniqueMode.RANDOM:
        length = kwargs.get("length", 8)

        def unique_key_fn(_: int) -> str:
            return "".join(random.choices(RAND_CHARS, k=length))

    elif mode == UniqueMode.INTEGER:
        length = kwargs.get("length", 3)

        def unique_key_fn(iteration: int) -> str:
            return str(iteration).rjust(length, "0")

    else:
        raise NotImplementedError

    if not force_unique and not os.path.isfile(path):
        return path

    directory, file_name = os.path.split(path)
    file_name, extension = os.path.splitext(file_name)

    for i in range(max_iterations):
        unique_key = unique_key_fn(i + 1)
        new_file_name = f"{file_name}{delimiter}{unique_key}{extension}"
        check_path = os.path.join(directory, new_file_name)
        if not os.path.isfile(check_path):
            return check_path


if __name__ == '__main__':
    import doctest
    doctest.testmod()



from typing import Optional, Sequence

from xappt.models.parameter.model import ParameterDescriptor
from xappt.models.parameter.validators import *


__all__ = [
    'ParamString',
    'ParamBool',
    'ParamInt',
    'ParamFloat',
    'ParamList',
]


class ParamString(ParameterDescriptor):
    def __init__(self, *, choices: Optional[Sequence[str]] = None, **kwargs):
        validators = [
            ValidateDefault,
            ValidateType,
            ValidateChoiceStr,
        ] + kwargs.get('validators', [])
        kwargs['validators'] = validators
        super().__init__(data_type=str, choices=choices, **kwargs)


class ParamBool(ParameterDescriptor):
    def __init__(self, **kwargs):
        validators = [
            ValidateBoolFromString,
            ValidateDefault,
            ValidateType,
        ] + kwargs.get('validators', [])
        kwargs['validators'] = validators
        super().__init__(data_type=bool, **kwargs)


class ParamInt(ParameterDescriptor):
    def __init__(self, *, minimum: Optional[int] = None, maximum: Optional[int] = None,
                 choices: Optional[Sequence[str]] = None, **kwargs):
        validators = [
            ValidateDefaultInt,
            ValidateChoiceInt,
            (ValidateRange, minimum, maximum),
        ] + kwargs.get('validators', [])
        kwargs['validators'] = validators
        options = kwargs.get("options", {})
        if minimum is not None:
            options['minimum'] = minimum
        if maximum is not None:
            options['maximum'] = maximum
        kwargs['options'] = options
        super().__init__(data_type=int, choices=choices, **kwargs)


class ParamFloat(ParameterDescriptor):
    def __init__(self, *, minimum: Optional[float] = None, maximum: Optional[float] = None, **kwargs):
        validators = [
            ValidateDefault,
            ValidateType,
            (ValidateRange, minimum, maximum),
        ] + kwargs.get('validators', [])
        kwargs['validators'] = validators
        options = kwargs.get("options", {})
        if minimum is not None:
            options['minimum'] = minimum
        if maximum is not None:
            options['maximum'] = maximum
        kwargs['options'] = options
        super().__init__(data_type=float, **kwargs)


class ParamList(ParameterDescriptor):
    def __init__(self, *, choices: Optional[Sequence[str]] = None, **kwargs):
        validators = [
            ValidateDefault,
            ValidateTypeList,
            ValidateChoiceList,
        ] + kwargs.get('validators', [])
        if choices is None:
            choices = []
        super().__init__(data_type=list, choices=choices, validators=validators, **kwargs)



from __future__ import annotations

import copy

from collections import defaultdict

from typing import Any, Callable, Dict, DefaultDict, List, Optional, Sequence, Set, Tuple, Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xappt.models.parameter.validators import BaseValidator


class ParameterCallback:
    def __init__(self):
        self._callback_functions: Set[Callable] = set()
        self._deferred_operations: DefaultDict[str, Set[Optional[Callable]]] = defaultdict(set)
        self._paused = False

    def add(self, cb: Callable):
        self._deferred_operations['add'].add(cb)

    def remove(self, cb: Callable):
        self._deferred_operations['remove'].add(cb)

    def clear(self):
        self._deferred_operations['clear'].add(None)

    def _run_deferred_ops(self):
        for operation, functions in self._deferred_operations.items():
            op_fn = getattr(self._callback_functions, operation)
            for function in functions:
                if function is None:
                    op_fn()
                else:
                    try:
                        op_fn(function)
                    except KeyError:
                        pass
        self._deferred_operations.clear()

    def invoke(self, sender: Parameter):
        self._run_deferred_ops()
        if self._paused:
            return
        for fn in self._callback_functions:
            fn(param=sender)

    @property
    def paused(self) -> bool:
        return self._paused

    @paused.setter
    def paused(self, value: bool):
        self._paused = value


class Parameter:
    def __init__(self, name: str, *, data_type: Type, **kwargs):
        self.name: str = name
        self.data_type: Type = data_type
        self.description: str = kwargs.get('description', "")
        self.default: Any = kwargs['default']
        self.required: bool = kwargs['required']
        self._choices: Optional[Sequence] = kwargs.get('choices')
        self._options: Dict[str: Any] = kwargs.get('options', {})
        self.validators: List[BaseValidator] = []
        self._value = kwargs['value']
        self.metadata: Dict = kwargs.get('metadata', {})

        for validator in kwargs.get('validators', []):
            if isinstance(validator, Tuple):
                validator, *args = validator
            else:
                args = []
            self.validators.append(validator(self, *args))

        self.on_value_changed = ParameterCallback()
        self.on_choices_changed = ParameterCallback()
        self.on_options_changed = ParameterCallback()

    def validate(self, value: Any) -> Any:
        for validator in self.validators:
            value = validator.validate(value)
        return value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.on_value_changed.invoke(self)

    @property
    def choices(self) -> Optional[Sequence]:
        return self._choices

    @choices.setter
    def choices(self, new_choices: Optional[Sequence]):
        self._choices = new_choices
        self.on_choices_changed.invoke(self)

    @property
    def options(self) -> Dict:
        return self._options

    def option(self, key: str, default: Any) -> Any:
        return self._options.get(key, default)

    def set_option(self, key: str, value: Any):
        self._options[key] = value
        self.on_options_changed.invoke(self)


class ParameterDescriptor:
    __counter = 0

    def __init__(self, data_type: Type, **kwargs):
        cls = self.__class__
        self.storage_name = '_{}#{}'.format(cls.__name__, cls.__counter)
        kwargs['data_type'] = data_type
        if 'default' not in kwargs:
            kwargs['required'] = True
        else:
            kwargs['required'] = False
        kwargs['default'] = kwargs.get('default')
        kwargs['value'] = kwargs.get('value', kwargs['default'])
        if "description" not in kwargs:
            kwargs['description'] = ""
        self.param_setup_args = copy.deepcopy(kwargs)
        cls.__counter += 1

    def get_parameter(self, instance):
        param = getattr(instance, self.storage_name, None)
        if param is None:
            # Create new parameter instance of descriptor properties
            param = Parameter(**self.param_setup_args)
            setattr(instance, self.storage_name, param)
            return param
        return param

    def __get__(self, instance, owner):
        if instance is None:
            # Call was not through an instance
            return self
        else:
            return self.get_parameter(instance)

    def __set__(self, instance, value):
        raise AttributeError("Parameter is read only. Use Parameter's value attribute")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import setuptools

from typing import List

from xappt.utilities import git_tools

REPOSITORY_PATH = os.path.dirname(os.path.abspath(__file__))
VERSION_PATH = os.path.join(REPOSITORY_PATH, "xappt", "__version__.py")

os.chdir(REPOSITORY_PATH)


def build_package_list(base_pkg: str, *, exclude: List[str] = None) -> List[str]:
    if exclude is None:
        exclude = []
    packages = [base_pkg]
    for package in setuptools.find_packages(base_pkg):
        pkg_name = f"{base_pkg}.{package}"
        if pkg_name in exclude:
            continue
        packages.append(pkg_name)
    return packages


def long_description() -> str:
    with open("README.md", "r", encoding="utf8") as fh:
        return fh.read()


def get_version() -> str:
    with open(VERSION_PATH, "r") as fp:
        version_contents = fp.read()

    loc = locals()
    exec(version_contents, {}, loc)
    __version__ = loc['__version__']
    assert __version__ is not None

    return __version__


def update_build(new_build: str):
    version = get_version()
    with open(VERSION_PATH, "w") as fp:
        fp.write(f'__version__ = "{version}"\n')
        fp.write(f'__build__ = "{new_build}"\n')


def requirements(variation=None):
    req_file = "requirements"
    if variation is not None:
        req_file = f"requirements-{variation}"
    req_path = os.path.join(os.path.dirname(__file__), f"{req_file}.txt")
    with open(req_path, "rb") as fp:
        for line in fp.readlines():
            line = line.decode('utf-8').strip()
            if len(line) == 0:
                continue
            if line.startswith("#"):
                continue
            yield line


def main():
    if git_tools.is_dirty(REPOSITORY_PATH):
        print("Local repository is not clean")
        while True:
            result = input("Proceed anyway? (y|n): ").lower()
            if result not in ("y", "n"):
                print(f"Invalid response '{result}', please try again.")
                continue
            if result == "n":
                return
            break

    setup_dict = {
        'name': 'xappt',
        'version': get_version(),
        'author': 'Christopher Montesano',
        'author_email': 'cmdev.00+xappt@gmail.com',
        'url': 'https://github.com/cmontesano/xappt',
        'description': 'Extensible Application Toolkit.',
        'long_description': long_description(),
        'long_description_content_type': 'text/markdown',
        'packages': build_package_list('xappt'),
        'include_package_data': True,
        'zip_safe': False,
        'license': 'MIT',
        'platforms': 'any',
        'python_requires': '>=3.7, <4',
        'install_requires': list(requirements()),
        'classifiers': [
            'Topic :: Utilities',
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
        ],
        'entry_points': {
            'console_scripts': ['xappt=xappt.cli:entry_point'],
        },
    }

    commit_id = git_tools.commit_id(REPOSITORY_PATH, short=True)

    update_build(commit_id)
    setuptools.setup(**setup_dict)
    update_build("dev")


if __name__ == '__main__':
    main()



class BasePlugin:
    @classmethod
    def name(cls) -> str:
        """ Use this to specify the name of your plugin. """
        return cls.__name__.lower()

    @classmethod
    def help(cls) -> str:
        """ Provide a short help string or description. """
        return ""

    @classmethod
    def collection(cls) -> str:
        """ Use this to group your plugin with other like plugins. """
        return ""
